@@ -1,34 +1,44 @@
+import collections
+import functools
 import hashlib
 import inspect
 import json
 import re
 import sys
+import types
 import typing
 import urllib.parse
 from abc import ABCMeta
 from collections import OrderedDict, defaultdict
-from collections.abc import Hashable, Iterable
 from decimal import Decimal
 from enum import Enum
-from typing import DefaultDict, Generic, List, Optional, Type, TypeVar, Union
+from typing import Any, DefaultDict, Generic, List, Optional, Tuple, Type, TypeVar, Union
 
 import inflection
 import uritemplate
 from django.apps import apps
+from django.db.models.constants import LOOKUP_SEP
+from django.db.models.fields.related_descriptors import (
+    ForwardManyToOneDescriptor, ManyToManyDescriptor, ReverseManyToOneDescriptor,
+    ReverseOneToOneDescriptor,
+)
 from django.db.models.fields.reverse_related import ForeignObjectRel
-from django.urls.resolvers import (  # type: ignore
+from django.db.models.sql.query import Query
+from django.urls.converters import get_converters
+from django.urls.resolvers import (  # type: ignore[attr-defined]
     _PATH_PARAMETER_COMPONENT_RE, RegexPattern, Resolver404, RoutePattern, URLPattern, URLResolver,
     get_resolver,
 )
 from django.utils.functional import Promise, cached_property
 from django.utils.module_loading import import_string
+from django.utils.translation import gettext_lazy as _
 from rest_framework import exceptions, fields, mixins, serializers, versioning
 from rest_framework.settings import api_settings
 from rest_framework.test import APIRequestFactory
 from rest_framework.utils.mediatypes import _MediaType
 from uritemplate import URITemplate
 
-from drf_spectacular.drainage import error, warn
+from drf_spectacular.drainage import Literal, _TypedDictMeta, cache, error, warn
 from drf_spectacular.settings import spectacular_settings
 from drf_spectacular.types import (
     DJANGO_PATH_CONVERTER_MAPPING, OPENAPI_TYPE_MAPPING, PYTHON_TYPE_MAPPING, OpenApiTypes,
@@ -41,6 +51,17 @@ except ImportError:
     class Choices:  # type: ignore
         pass
 
+# types.UnionType was added in Python 3.10 for new PEP 604 pipe union syntax
+if hasattr(types, 'UnionType'):
+    UNION_TYPES: Tuple[Any, ...] = (typing.Union, types.UnionType)  # type: ignore
+else:
+    UNION_TYPES = (typing.Union,)
+
+if sys.version_info >= (3, 8):
+    CACHED_PROPERTY_FUNCS = (functools.cached_property, cached_property)  # type: ignore
+else:
+    CACHED_PROPERTY_FUNCS = (cached_property,)  # type: ignore
+
 T = TypeVar('T')
 
 
@@ -48,10 +69,6 @@ class UnableToProceedError(Exception):
     pass
 
 
-def anyisinstance(obj, type_list):
-    return any([isinstance(obj, t) for t in type_list])
-
-
 def get_class(obj) -> type:
     return obj if inspect.isclass(obj) else obj.__class__
 
@@ -77,6 +94,10 @@ def is_list_serializer(obj) -> bool:
     return isinstance(force_instance(obj), serializers.ListSerializer)
 
 
+def is_basic_serializer(obj) -> bool:
+    return is_serializer(obj) and not is_list_serializer(obj)
+
+
 def is_field(obj):
     # make sure obj is a serializer field and nothing else.
     # guard against serializers because BaseSerializer(Field)
@@ -84,13 +105,38 @@ def is_field(obj):
 
 
 def is_basic_type(obj, allow_none=True):
-    if not isinstance(obj, Hashable):
+    if not isinstance(obj, collections.abc.Hashable):
         return False
     if not allow_none and (obj is None or obj is OpenApiTypes.NONE):
         return False
-    return obj in OPENAPI_TYPE_MAPPING or obj in PYTHON_TYPE_MAPPING
+    return obj in get_openapi_type_mapping() or obj in PYTHON_TYPE_MAPPING
 
 
+def is_patched_serializer(serializer, direction):
+    return bool(
+        spectacular_settings.COMPONENT_SPLIT_PATCH
+        and serializer.partial
+        and not serializer.read_only
+        and not (spectacular_settings.COMPONENT_SPLIT_REQUEST and direction == 'response')
+    )
+
+
+def is_trivial_string_variation(a: str, b: str):
+    a = (a or '').strip().lower().replace(' ', '_').replace('-', '_')
+    b = (b or '').strip().lower().replace(' ', '_').replace('-', '_')
+    return a == b
+
+
+def assert_basic_serializer(serializer):
+    assert is_basic_serializer(serializer), (
+        f'internal assumption violated because we expected a basic serializer here and '
+        f'instead got a "{serializer}". This may be the result of another app doing '
+        f'some unexpected magic or an invalid internal call. Feel free to report this '
+        f'as a bug at https://github.com/tfranzel/drf-spectacular/issues'
+    )
+
+
+@cache
 def get_lib_doc_excludes():
     # do not import on package level due to potential import recursion when loading
     # extensions as recommended:  USER's settings.py -> USER EXTENSIONS -> extensions.py
@@ -105,7 +151,7 @@ def get_lib_doc_excludes():
     ]
 
 
-def get_view_model(view):
+def get_view_model(view, emit_warnings=True):
     """
     obtain model from view via view's queryset. try safer view attribute first
     before going through get_queryset(), which may perform arbitrary operations.
@@ -118,18 +164,23 @@ def get_view_model(view):
     try:
         return view.get_queryset().model
     except Exception as exc:
-        warn(
-            f'failed to obtain model through view\'s queryset due to raised exception. '
-            f'prevent this either by setting "queryset = Model.objects.none()" on the view, '
-            f'having an empty fallback in get_queryset() or by using @extend_schema. '
-            f'(Exception: {exc})'
-        )
+        if emit_warnings:
+            warn(
+                f'Failed to obtain model through view\'s queryset due to raised exception. '
+                f'Prevent this either by setting "queryset = Model.objects.none()" on the '
+                f'view, checking for "getattr(self, "swagger_fake_view", False)" in '
+                f'get_queryset() or by simply using @extend_schema. (Exception: {exc})'
+            )
 
 
 def get_doc(obj):
     """ get doc string with fallback on obj's base classes (ignoring DRF documentation). """
+    def post_cleanup(doc: str):
+        # also clean up trailing whitespace for each line
+        return '\n'.join(line.rstrip() for line in doc.rstrip().split('\n'))
+
     if not inspect.isclass(obj):
-        return inspect.getdoc(obj) or ''
+        return post_cleanup(inspect.getdoc(obj) or '')
 
     def safe_index(lst, item):
         try:
@@ -142,26 +193,48 @@ def get_doc(obj):
     )
     for cls in obj.__mro__[:lib_barrier]:
         if cls.__doc__:
-            # dedent and cleanup of trailing white space on each line
-            doc = inspect.cleandoc(cls.__doc__)
-            doc = '\n'.join(line.rstrip() for line in doc.rstrip().split('\n'))
-            return doc
+            return post_cleanup(inspect.cleandoc(cls.__doc__))
     return ''
 
 
+def get_type_hints(obj):
+    """ unpack wrapped partial object and use actual func object """
+    if isinstance(obj, functools.partial):
+        obj = obj.func
+    return typing.get_type_hints(obj)
+
+
+@cache
+def get_openapi_type_mapping():
+    return {
+        **OPENAPI_TYPE_MAPPING,
+        OpenApiTypes.OBJECT: build_generic_type(),
+    }
+
+
+def build_generic_type():
+    if spectacular_settings.GENERIC_ADDITIONAL_PROPERTIES is None:
+        return {'type': 'object'}
+    elif spectacular_settings.GENERIC_ADDITIONAL_PROPERTIES == 'bool':
+        return {'type': 'object', 'additionalProperties': True}
+    else:
+        return {'type': 'object', 'additionalProperties': {}}
+
+
 def build_basic_type(obj):
     """
     resolve either enum or actual type and yield schema template for modification
     """
-    if obj in OPENAPI_TYPE_MAPPING:
-        return dict(OPENAPI_TYPE_MAPPING[obj])
+    openapi_type_mapping = get_openapi_type_mapping()
+    if obj is None or type(obj) is None or obj is OpenApiTypes.NONE:
+        return None
+    elif obj in openapi_type_mapping:
+        return dict(openapi_type_mapping[obj])
     elif obj in PYTHON_TYPE_MAPPING:
-        return dict(OPENAPI_TYPE_MAPPING[PYTHON_TYPE_MAPPING[obj]])
-    elif obj is None or type(obj) is None:
-        return dict(OPENAPI_TYPE_MAPPING[OpenApiTypes.NONE])
+        return dict(openapi_type_mapping[PYTHON_TYPE_MAPPING[obj]])
     else:
         warn(f'could not resolve type for "{obj}". defaulting to "string"')
-        return dict(OPENAPI_TYPE_MAPPING[OpenApiTypes.STR])
+        return dict(openapi_type_mapping[OpenApiTypes.STR])
 
 
 def build_array_type(schema, min_length=None, max_length=None):
@@ -228,7 +301,10 @@ def build_parameter_type(
         deprecated=False,
         explode=None,
         style=None,
+        default=None,
+        allow_blank=True,
         examples=None,
+        extensions=None,
 ):
     irrelevant_field_meta = ['readOnly', 'writeOnly']
     if location == OpenApiParameter.PATH:
@@ -250,8 +326,14 @@ def build_parameter_type(
         schema['style'] = style
     if enum:
         schema['schema']['enum'] = sorted(enum)
+    if default is not None and 'default' not in irrelevant_field_meta:
+        schema['schema']['default'] = default
+    if not allow_blank and schema['schema'].get('type') == 'string':
+        schema['schema']['minLength'] = schema['schema'].get('minLength', 1)
     if examples:
         schema['examples'] = examples
+    if extensions:
+        schema.update(sanitize_specification_extensions(extensions))
     return schema
 
 
@@ -290,16 +372,46 @@ def build_choice_field(field):
     return schema
 
 
-def build_root_object(paths, components):
+def build_bearer_security_scheme_object(header_name, token_prefix, bearer_format=None):
+    """ Either build a bearer scheme or a fallback due to OpenAPI 3.0.3 limitations """
+    # normalize Django header quirks
+    if header_name.startswith('HTTP_'):
+        header_name = header_name[5:]
+    header_name = header_name.replace('_', '-').capitalize()
+
+    if token_prefix == 'Bearer' and header_name == 'Authorization':
+        return {
+            'type': 'http',
+            'scheme': 'bearer',
+            **({'bearerFormat': bearer_format} if bearer_format else {}),
+        }
+    else:
+        return {
+            'type': 'apiKey',
+            'in': 'header',
+            'name': header_name,
+            'description': _(
+                'Token-based authentication with required prefix "%s"'
+            ) % token_prefix
+        }
+
+
+def build_root_object(paths, components, version):
     settings = spectacular_settings
+    if settings.VERSION and version:
+        version = f'{settings.VERSION} ({version})'
+    else:
+        version = settings.VERSION or version or ''
     root = {
         'openapi': '3.0.3',
         'info': {
             'title': settings.TITLE,
-            'version': settings.VERSION,
+            'version': version,
+            **sanitize_specification_extensions(settings.EXTENSIONS_INFO),
         },
         'paths': {**paths, **settings.APPEND_PATHS},
-        'components': components
+        'components': components,
+        **sanitize_specification_extensions(settings.EXTENSIONS_ROOT),
     }
     if settings.DESCRIPTION:
         root['info']['description'] = settings.DESCRIPTION
@@ -342,27 +454,44 @@ def _follow_field_source(model, path: List[str]):
         # end of traversal
         if isinstance(field_or_property, property):
             return field_or_property.fget
-        elif isinstance(field_or_property, cached_property):
+        elif isinstance(field_or_property, CACHED_PROPERTY_FUNCS):
             return field_or_property.func
         elif callable(field_or_property):
             return field_or_property
+        elif isinstance(field_or_property, ManyToManyDescriptor):
+            if field_or_property.reverse:
+                return field_or_property.rel.target_field  # m2m reverse
+            else:
+                return field_or_property.field.target_field  # m2m forward
+        elif isinstance(field_or_property, ReverseOneToOneDescriptor):
+            return field_or_property.related.target_field  # o2o reverse
+        elif isinstance(field_or_property, ReverseManyToOneDescriptor):
+            return field_or_property.rel.target_field  # type: ignore # foreign reverse
+        elif isinstance(field_or_property, ForwardManyToOneDescriptor):
+            return field_or_property.field.target_field  # type: ignore # o2o & foreign forward
         else:
             field = model._meta.get_field(path[0])
             if isinstance(field, ForeignObjectRel):
-                # resolve DRF internal object to PK field as approximation
-                return field.get_related_field()  # type: ignore
+                # case only occurs when relations are traversed in reverse and
+                # not via the related_name (default: X_set) but the model name.
+                return field.target_field
             else:
                 return field
     else:
-        if isinstance(field_or_property, property) or callable(field_or_property):
+        if (
+            isinstance(field_or_property, (property,) + CACHED_PROPERTY_FUNCS)
+            or callable(field_or_property)
+        ):
             if isinstance(field_or_property, property):
-                target_model = field_or_property.fget.__annotations__.get('return')
+                target_model = _follow_return_type(field_or_property.fget)
+            elif isinstance(field_or_property, CACHED_PROPERTY_FUNCS):
+                target_model = _follow_return_type(field_or_property.func)
             else:
-                target_model = field_or_property.__annotations__.get('return')
+                target_model = _follow_return_type(field_or_property)
             if not target_model:
                 raise UnableToProceedError(
                     f'could not follow field source through intermediate property "{path[0]}" '
-                    f'on model {model}. please add a type hint on the model\'s property/function '
+                    f'on model {model}. Please add a type hint on the model\'s property/function '
                     f'to enable traversal of the source path "{".".join(path)}".'
                 )
             return _follow_field_source(target_model, path[1:])
@@ -371,7 +500,26 @@ def _follow_field_source(model, path: List[str]):
             return _follow_field_source(target_model, path[1:])
 
 
-def follow_field_source(model, path):
+def _follow_return_type(a_callable):
+    target_type = get_type_hints(a_callable).get('return')
+    if target_type is None:
+        return target_type
+    origin, args = _get_type_hint_origin(target_type)
+    if origin in UNION_TYPES:
+        type_args = [arg for arg in args if arg is not type(None)]  # noqa: E721
+        if len(type_args) > 1:
+            warn(
+                f'could not traverse Union type, because we don\'t know which type to choose '
+                f'from {type_args}. Consider terminating "source" on a custom property '
+                f'that indicates the expected Optional/Union type. Defaulting to "string"'
+            )
+            return target_type
+        # Optional:
+        return type_args[0]
+    return target_type
+
+
+def follow_field_source(model, path, emit_warnings=True):
     """
     a model traversal chain "foreignkey.foreignkey.value" can either end with an actual model field
     instance "value" or a model property function named "value". differentiate the cases.
@@ -381,19 +529,32 @@ def follow_field_source(model, path):
     try:
         return _follow_field_source(model, path)
     except UnableToProceedError as e:
-        warn(e)
+        if emit_warnings:
+            warn(e)
     except Exception as exc:
-        warn(
-            f'could not resolve field on model {model} with path "{".".join(path)}". '
-            f'this is likely a custom field that does some unknown magic. maybe '
-            f'consider annotating the field/property? defaulting to "string". (Exception: {exc})'
-        )
+        if emit_warnings:
+            warn(
+                f'could not resolve field on model {model} with path "{".".join(path)}". '
+                f'This is likely a custom field that does some unknown magic. Maybe '
+                f'consider annotating the field/property? Defaulting to "string". (Exception: {exc})'
+            )
 
     def dummy_property(obj) -> str:
-        pass
+        pass  # pragma: no cover
     return dummy_property
 
 
+def follow_model_field_lookup(model, lookup):
+    """
+    Follow a model lookup `foreignkey__foreignkey__field` in the same
+    way that Django QuerySet.filter() does, returning the final models.Field.
+    """
+    query = Query(model)
+    lookup_splitted = lookup.split(LOOKUP_SEP)
+    _, field, _, _ = query.names_to_path(lookup_splitted, query.get_meta())
+    return field
+
+
 def alpha_operation_sorter(endpoint):
     """ sort endpoints first alphanumerically by path, then by method order """
     path, path_regex, method, callback = endpoint
@@ -432,7 +593,7 @@ class ResolvedComponent:
 
     @property
     def ref(self) -> dict:
-        assert self.name and self.type and self.object
+        assert self.__bool__()
         return {'$ref': f'#/components/{self.type}/{self.name}'}
 
 
@@ -556,6 +717,7 @@ def deep_import_string(string):
         pass
 
 
+@cache
 def load_enum_name_overrides():
     overrides = {}
     for name, choices in spectacular_settings.ENUM_NAME_OVERRIDES.items():
@@ -571,14 +733,20 @@ def load_enum_name_overrides():
             choices = choices.choices
         if inspect.isclass(choices) and issubclass(choices, Enum):
             choices = [c.value for c in choices]
-        if isinstance(choices, Iterable) and isinstance(choices[0], str):
-            choices = [(c, c) for c in choices]
-        overrides[list_hash(list(dict(choices).keys()))] = name
+        normalized_choices = []
+        for choice in choices:
+            if isinstance(choice, str):
+                normalized_choices.append((choice, choice))  # simple choice list
+            elif isinstance(choice[1], (list, tuple)):
+                normalized_choices.extend(choice[1])  # categorized nested choices
+            else:
+                normalized_choices.append(choice)  # normal 2-tuple form
+        overrides[list_hash(list(dict(normalized_choices).keys()))] = name
 
     if len(spectacular_settings.ENUM_NAME_OVERRIDES) != len(overrides):
         error(
-            'ENUM_NAME_OVERRIDES has duplication issues. encountered multiple names '
-            'for the same choice set. enum naming might be unexpected.'
+            'ENUM_NAME_OVERRIDES has duplication issues. Encountered multiple names '
+            'for the same choice set. Enum naming might be unexpected.'
         )
     return overrides
 
@@ -587,16 +755,36 @@ def list_hash(lst):
     return hashlib.sha256(json.dumps(list(lst), sort_keys=True).encode()).hexdigest()
 
 
-def resolve_regex_path_parameter(path_regex, variable, available_formats):
+def anchor_pattern(pattern: str) -> str:
+    if not pattern.startswith('^'):
+        pattern = '^' + pattern
+    if not pattern.endswith('$'):
+        pattern = pattern + '$'
+    return pattern
+
+
+def resolve_django_path_parameter(path_regex, variable, available_formats):
     """
     convert django style path parameters to OpenAPI parameters.
-    TODO also try to handle regular grouped regex parameters
     """
+    registered_converters = get_converters()
     for match in _PATH_PARAMETER_COMPONENT_RE.finditer(path_regex):
         converter, parameter = match.group('converter'), match.group('parameter')
         enum_values = None
 
-        if converter and converter.startswith('drf_format_suffix_'):
+        if api_settings.SCHEMA_COERCE_PATH_PK and parameter == 'pk':
+            parameter = 'id'
+        elif spectacular_settings.SCHEMA_COERCE_PATH_PK_SUFFIX and parameter.endswith('_pk'):
+            parameter = f'{parameter[:-3]}_id'
+
+        if parameter != variable:
+            continue
+        # RE also matches untyped patterns (e.g. "<id>")
+        if not converter:
+            return None
+
+        # special handling for drf_format_suffix
+        if converter.startswith('drf_format_suffix_'):
             explicit_formats = converter[len('drf_format_suffix_'):].split('_')
             enum_values = [
                 f'.{suffix}' for suffix in explicit_formats if suffix in available_formats
@@ -605,16 +793,63 @@ def resolve_regex_path_parameter(path_regex, variable, available_formats):
         elif converter == 'drf_format_suffix':
             enum_values = [f'.{suffix}' for suffix in available_formats]
 
+        if converter in spectacular_settings.PATH_CONVERTER_OVERRIDES:
+            override = spectacular_settings.PATH_CONVERTER_OVERRIDES[converter]
+            if is_basic_type(override):
+                schema = build_basic_type(override)
+            elif isinstance(override, dict):
+                schema = dict(override)
+            else:
+                warn(
+                    f'Unable to use path converter override for "{converter}". '
+                    f'Please refer to the documentation on how to use this.'
+                )
+                return None
+        elif converter in DJANGO_PATH_CONVERTER_MAPPING:
+            schema = build_basic_type(DJANGO_PATH_CONVERTER_MAPPING[converter])
+        elif converter in registered_converters:
+            # gracious fallback for custom converters that have no override specified.
+            schema = build_basic_type(OpenApiTypes.STR)
+            schema['pattern'] = anchor_pattern(registered_converters[converter].regex)
+        else:
+            error(f'Encountered path converter "{converter}" that is unknown to Django.')
+            return None
+
+        return build_parameter_type(
+            name=variable,
+            schema=schema,
+            location=OpenApiParameter.PATH,
+            enum=enum_values,
+        )
+
+    return None
+
+
+def resolve_regex_path_parameter(path_regex, variable):
+    """
+    convert regex path parameter to OpenAPI parameter, if pattern is
+    explicitly chosen and not the generic non-empty default '[^/.]+'.
+    """
+    for parameter, pattern in analyze_named_regex_pattern(path_regex).items():
         if api_settings.SCHEMA_COERCE_PATH_PK and parameter == 'pk':
             parameter = 'id'
+        elif spectacular_settings.SCHEMA_COERCE_PATH_PK_SUFFIX and parameter.endswith('_pk'):
+            parameter = f'{parameter[:-3]}_id'
 
-        if parameter == variable and converter in DJANGO_PATH_CONVERTER_MAPPING:
-            return build_parameter_type(
-                name=parameter,
-                schema=build_basic_type(DJANGO_PATH_CONVERTER_MAPPING[converter]),
-                location=OpenApiParameter.PATH,
-                enum=enum_values,
-            )
+        if parameter != variable:
+            continue
+        # do not use default catch-all pattern and defer to model resolution
+        if pattern == '[^/.]+':
+            return None
+
+        return build_parameter_type(
+            name=variable,
+            schema={
+                **build_basic_type(OpenApiTypes.STR),
+                'pattern': anchor_pattern(pattern),
+            },
+            location=OpenApiParameter.PATH,
+        )
 
     return None
 
@@ -656,7 +891,7 @@ def modify_for_versioning(patterns, method, path, view, requested_version):
                 urlconf=tuple(detype_pattern(p) for p in patterns)
             ).resolve(path)
         except Resolver404:
-            error(f"namespace versioning path resolution failed for {path}. path will be ignored.")
+            error(f"namespace versioning path resolution failed for {path}. Path will be ignored.")
     elif issubclass(view.versioning_class, versioning.AcceptHeaderVersioning):
         # Append the version into request accepted_media_type.
         # e.g "application/json; version=1.0"
@@ -673,6 +908,51 @@ def modify_for_versioning(patterns, method, path, view, requested_version):
     return path
 
 
+def analyze_named_regex_pattern(path):
+    """ safely extract named groups and their pattern from given regex pattern """
+    result = {}
+    stack = 0
+    name_capture, name_buffer = False, ''
+    regex_capture, regex_buffer = False, ''
+    i = 0
+    while i < len(path):
+        # estimate state at position i
+        skip = False
+        if path[i] == '\\':
+            ff = 2
+        elif path[i:i + 4] == '(?P<':
+            skip = True
+            name_capture = True
+            ff = 4
+        elif path[i] in '(':
+            stack += 1
+            ff = 1
+        elif path[i] == '>' and name_capture:
+            assert name_buffer
+            name_capture = False
+            regex_capture = True
+            skip = True
+            ff = 1
+        elif path[i] in ')':
+            if regex_capture and not stack:
+                regex_capture = False
+                result[name_buffer] = regex_buffer
+                name_buffer, regex_buffer = '', ''
+            else:
+                stack -= 1
+            ff = 1
+        else:
+            ff = 1
+        # fill buffer based on state
+        if name_capture and not skip:
+            name_buffer += path[i:i + ff]
+        elif regex_capture and not skip:
+            regex_buffer += path[i:i + ff]
+        i += ff
+    assert not stack
+    return result
+
+
 def detype_pattern(pattern):
     """
     return an equivalent pattern that accepts arbitrary values for path parameters.
@@ -701,8 +981,14 @@ def detype_pattern(pattern):
             is_endpoint=pattern._is_endpoint
         )
     elif isinstance(pattern, RegexPattern):
+        detyped_regex = pattern._regex
+        for name, regex in analyze_named_regex_pattern(pattern._regex).items():
+            detyped_regex = detyped_regex.replace(
+                f'(?P<{name}>{regex})',
+                f'(?P<{name}>[^/]+)',
+            )
         return RegexPattern(
-            regex=re.sub(r'\(\?P<(\w+)>.+?\)', r'(?P<\1>[^/]+)', pattern._regex),
+            regex=detyped_regex,
             name=pattern.name,
             is_endpoint=pattern._is_endpoint
         )
@@ -742,6 +1028,17 @@ def sanitize_result_object(result):
     return result
 
 
+def sanitize_specification_extensions(extensions):
+    # https://spec.openapis.org/oas/v3.0.3#specification-extensions
+    output = {}
+    for key, value in extensions.items():
+        if not re.match(r'^x-', key):
+            warn(f'invalid extension {key!r}. vendor extensions must start with "x-"')
+        else:
+            output[key] = value
+    return output
+
+
 def camelize_operation(path, operation):
     for path_variable in re.findall(r'\{(\w+)\}', path):
         path = path.replace(
@@ -759,15 +1056,29 @@ def camelize_operation(path, operation):
 
 
 def build_mock_request(method, path, view, original_request, **kwargs):
+    """ build a mocked request and use original request as reference if available """
     request = getattr(APIRequestFactory(), method.lower())(path=path)
     request = view.initialize_request(request)
+    if original_request:
+        request.user = original_request.user
+        request.auth = original_request.auth
+        # ignore headers related to authorization as it has been handled above.
+        # also ignore ACCEPT as the MIME type refers to SpectacularAPIView and the
+        # version (if available) has already been processed by SpectacularAPIView.
+        for name, value in original_request.META.items():
+            if not name.startswith('HTTP_'):
+                continue
+            if name in ['HTTP_ACCEPT', 'HTTP_COOKIE', 'HTTP_AUTHORIZATION']:
+                continue
+            request.META[name] = value
     return request
 
 
 def set_query_parameters(url, **kwargs) -> str:
     """ deconstruct url, safely attach query parameters in kwargs, and serialize again """
     scheme, netloc, path, params, query, fragment = urllib.parse.urlparse(url)
-    query = {k: v for k, v in kwargs.items() if v is not None}
+    query = urllib.parse.parse_qs(query)
+    query.update({k: v for k, v in kwargs.items() if v is not None})
     query = urllib.parse.urlencode(query, doseq=True)
     return urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))
 
@@ -800,9 +1111,16 @@ def resolve_type_hint(hint):
 
     if origin is None and is_basic_type(hint, allow_none=False):
         return build_basic_type(hint)
+    elif origin is None and inspect.isclass(hint) and issubclass(hint, tuple):
+        # a convoluted way to catch NamedTuple. suggestions welcome.
+        if get_type_hints(hint):
+            properties = {k: resolve_type_hint(v) for k, v in get_type_hints(hint).items()}
+        else:
+            properties = {k: build_basic_type(OpenApiTypes.ANY) for k in hint._fields}
+        return build_object_type(properties=properties, required=properties.keys())
     elif origin is list or hint is list:
         return build_array_type(
-            resolve_type_hint(args[0]) if args else build_basic_type(OpenApiTypes.OBJECT)
+            resolve_type_hint(args[0]) if args else build_basic_type(OpenApiTypes.ANY)
         )
     elif origin is tuple:
         return build_array_type(
@@ -819,24 +1137,44 @@ def resolve_type_hint(hint):
         return build_array_type(resolve_type_hint(args[0]))
     elif origin is frozenset:
         return build_array_type(resolve_type_hint(args[0]))
-    elif hasattr(typing, 'Literal') and origin is typing.Literal:
-        # python >= 3.8
+    elif origin is Literal:
+        # Literal only works for python >= 3.8 despite typing_extensions, because it
+        # behaves slightly different w.r.t. __origin__
         schema = {'enum': list(args)}
         if all(type(args[0]) is type(choice) for choice in args):
             schema.update(build_basic_type(type(args[0])))
         return schema
-    elif hasattr(typing, 'TypedDict') and isinstance(hint, typing._TypedDictMeta):
+    elif inspect.isclass(hint) and issubclass(hint, Enum):
+        schema = {'enum': [item.value for item in hint]}
+        mixin_base_types = [t for t in hint.__mro__ if is_basic_type(t)]
+        if mixin_base_types:
+            schema.update(build_basic_type(mixin_base_types[0]))
+        return schema
+    elif isinstance(hint, _TypedDictMeta):
         return build_object_type(
             properties={
-                k: resolve_type_hint(v) for k, v in typing.get_type_hints(hint).items()
+                k: resolve_type_hint(v) for k, v in get_type_hints(hint).items()
             }
         )
-    elif origin is typing.Union and len(args) == 2 and isinstance(None, args[1]):
-        # Optional[*] is resolved to Union[*, None]
-        schema = resolve_type_hint(args[0])
-        schema['nullable'] = True
+    elif origin in UNION_TYPES:
+        type_args = [arg for arg in args if arg is not type(None)]  # noqa: E721
+        if len(type_args) > 1:
+            schema = {'oneOf': [resolve_type_hint(arg) for arg in type_args]}
+        else:
+            schema = resolve_type_hint(type_args[0])
+        if type(None) in args:
+            schema['nullable'] = True
         return schema
-    elif origin is typing.Union:
-        return {'oneOf': [resolve_type_hint(arg) for arg in args]}
+    elif origin is collections.abc.Iterable:
+        return build_array_type(resolve_type_hint(args[0]))
     else:
         raise UnableToProceedError()
+
+
+def whitelisted(obj: object, classes: List[Type[object]], exact=False):
+    if not classes:
+        return True
+    if exact:
+        return obj.__class__ in classes
+    else:
+        return isinstance(obj, tuple(classes))
