@@ -1,17 +1,22 @@
-import functools
 import inspect
-from typing import Any, Dict, List, Optional, Type, Union
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union
 
-from rest_framework.fields import empty
+from rest_framework.fields import Field, empty
 from rest_framework.serializers import Serializer
 from rest_framework.settings import api_settings
 
-from drf_spectacular.drainage import error, get_view_methods, set_override, warn
+from drf_spectacular.drainage import (
+    Final, Literal, error, get_view_method_names, isolate_view_method, set_override, warn,
+)
+from drf_spectacular.types import OpenApiTypes, _KnownPythonTypes
 
-SerializerType = Union[Serializer, Type[Serializer]]
+_SerializerType = Union[Serializer, Type[Serializer]]
+_FieldType = Union[Field, Type[Field]]
+_ParameterLocationType = Literal['query', 'path', 'header', 'cookie']
+Direction = Literal['request', 'response']
 
 
-class PolymorphicProxySerializer:
+class PolymorphicProxySerializer(Serializer):
     """
     This class is to be used with :func:`@extend_schema <.extend_schema>` to
     signal a request/response might be polymorphic (accepts/returns data
@@ -31,16 +36,17 @@ class PolymorphicProxySerializer:
         def create(self, request, *args, **kwargs):
             return Response(...)
 
-    Beware that this is not a real serializer and therefore is not derived from
-    serializers.Serializer. It *cannot* be used in views as `serializer_class`
-    or as field in a actual serializer. You likely want to handle this in the
-    view method.
+    **Beware** that this is not a real serializer and it will raise an AssertionError
+    if used in that way. It **cannot** be used in views as `serializer_class`
+    or as field in an actual serializer. It is solely meant for annotation purposes.
 
     Also make sure that each sub-serializer has a field named after the value of
     ``resource_type_field_name`` (discriminator field). Generated clients will likely
-    depend on the existence of this field.
+    depend on the existence of this field. Setting ``resource_type_field_name``
+    to ``None`` will remove the discriminator altogether. This may be useful in
+    certain situations, but will most likely break client generation.
 
-    For that reason, it is **strongly** recommended to pass the ``Serializers`` as **list**,
+    It is **strongly** recommended to pass the ``Serializers`` as **list**,
     and by that let *drf-spectacular* retrieve the field and handle the mapping
     automatically. In special circumstances, the field may not available when
     drf-spectacular processes the serializer. In those cases you can explicitly state
@@ -50,14 +56,30 @@ class PolymorphicProxySerializer:
     def __init__(
             self,
             component_name: str,
-            serializers: Union[List[SerializerType], Dict[str, SerializerType]],
-            resource_type_field_name: str
+            serializers: Union[List[_SerializerType], Dict[str, _SerializerType]],
+            resource_type_field_name: Optional[str],
+            many: bool = False,
     ):
         self.component_name = component_name
         self.serializers = serializers
         self.resource_type_field_name = resource_type_field_name
-        self.partial = False
-        self.read_only = False
+        super().__init__(many=many)
+
+    @property
+    def data(self):
+        self._trap()
+
+    def to_internal_value(self, data):
+        self._trap()
+
+    def to_representation(self, instance):
+        self._trap()
+
+    def _trap(self):
+        raise AssertionError(
+            "PolymorphicProxySerializer is an annotation helper and not supposed to "
+            "be used for real requests. See documentation for correct usage."
+        )
 
 
 class OpenApiSchemaBase:
@@ -66,12 +88,13 @@ class OpenApiSchemaBase:
 
 class OpenApiExample(OpenApiSchemaBase):
     """
-    https://swagger.io/specification/
-     * Specification
-        * Schema
-           * ExampleObject
-    """
+    Helper class to document a API parameter / request body / response body
+    with a concrete example value.
 
+    The example will be attached to the operation object where appropriate,
+    i. e. where the given ``media_type``, ``status_code`` and modifiers match.
+    Example that do not match any scenario are ignored.
+    """
     def __init__(
             self,
             name: str,
@@ -81,8 +104,9 @@ class OpenApiExample(OpenApiSchemaBase):
             description: str = '',
             request_only: bool = False,
             response_only: bool = False,
+            parameter_only: Optional[Tuple[str, _ParameterLocationType]] = None,
             media_type: str = 'application/json',
-            status_codes: Optional[List[str]] = None
+            status_codes: Optional[List[str]] = None,
     ):
         self.name = name
         self.summary = summary
@@ -91,29 +115,44 @@ class OpenApiExample(OpenApiSchemaBase):
         self.external_value = external_value
         self.request_only = request_only
         self.response_only = response_only
+        self.parameter_only = parameter_only
         self.media_type = media_type
         self.status_codes = status_codes or ['200', '201']
 
 
 class OpenApiParameter(OpenApiSchemaBase):
-    QUERY = 'query'
-    PATH = 'path'
-    HEADER = 'header'
-    COOKIE = 'cookie'
+    """
+    Helper class to document request query/path/header/cookie parameters.
+    Can also be used to document response headers.
+
+    Please note that not all arguments apply to all ``location``/``type``/direction
+    variations, e.g. path parameters are ``required=True`` by definition.
+
+    For valid ``style`` choices please consult the
+    `OpenAPI specification <https://swagger.io/specification/#style-values>`_.
+    """
+    QUERY: Final = 'query'
+    PATH: Final = 'path'
+    HEADER: Final = 'header'
+    COOKIE: Final = 'cookie'
 
     def __init__(
             self,
-            name,
-            type=str,
-            location=QUERY,
-            required=False,
-            description='',
-            enum=None,
-            deprecated=False,
-            style=None,
-            explode=None,
+            name: str,
+            type: Union[_SerializerType, _KnownPythonTypes, OpenApiTypes, dict] = str,
+            location: _ParameterLocationType = QUERY,
+            required: bool = False,
+            description: str = '',
+            enum: Optional[List[Any]] = None,
+            deprecated: bool = False,
+            style: Optional[str] = None,
+            explode: Optional[bool] = None,
+            default: Any = None,
+            allow_blank: bool = True,
             examples: Optional[List[OpenApiExample]] = None,
-            exclude=False,
+            extensions: Optional[Dict[str, Any]] = None,
+            exclude: bool = False,
+            response: Union[bool, List[Union[int, str]]] = False,
     ):
         self.name = name
         self.type = type
@@ -124,29 +163,58 @@ class OpenApiParameter(OpenApiSchemaBase):
         self.deprecated = deprecated
         self.style = style
         self.explode = explode
+        self.default = default
+        self.allow_blank = allow_blank
         self.examples = examples or []
+        self.extensions = extensions
         self.exclude = exclude
+        self.response = response
+
+
+class OpenApiResponse(OpenApiSchemaBase):
+    """
+    Helper class to bundle a response object (``Serializer``, ``OpenApiType``,
+    raw schema, etc) together with a response object description and/or examples.
+    Examples can alternatively be provided via :func:`@extend_schema <.extend_schema>`.
+
+    This class is especially helpful for explicitly describing status codes on a
+    "Response Object" level.
+    """
+    def __init__(
+            self,
+            response: Any = None,
+            description: str = '',
+            examples: Optional[List[OpenApiExample]] = None
+    ):
+        self.response = response
+        self.description = description
+        self.examples = examples or []
+
+
+F = TypeVar('F', bound=Callable[..., Any])
 
 
 def extend_schema(
-        operation_id=None,
-        parameters=None,
-        request=empty,
-        responses=empty,
-        auth=None,
-        description=None,
-        summary=None,
-        deprecated=None,
-        tags=None,
-        exclude=False,
-        operation=None,
-        methods=None,
-        versions=None,
+        operation_id: Optional[str] = None,
+        parameters: Optional[List[Union[OpenApiParameter, _SerializerType]]] = None,
+        request: Any = empty,
+        responses: Any = empty,
+        auth: Optional[List[str]] = None,
+        description: Optional[str] = None,
+        summary: Optional[str] = None,
+        deprecated: Optional[bool] = None,
+        tags: Optional[List[str]] = None,
+        filters: Optional[bool] = None,
+        exclude: bool = False,
+        operation: Optional[Dict] = None,
+        methods: Optional[List[str]] = None,
+        versions: Optional[List[str]] = None,
         examples: Optional[List[OpenApiExample]] = None,
-):
+        extensions: Optional[Dict[str, Any]] = None,
+) -> Callable[[F], F]:
     """
-    decorator mainly for the "view" method kind. partially or completely overrides
-    what would be generated by drf-spectacular.
+    Decorator mainly for the "view" method kind. Partially or completely overrides
+    what would be otherwise generated by drf-spectacular.
 
     :param operation_id: replaces the auto-generated operation_id. make sure there
         are no naming collisions.
@@ -157,25 +225,44 @@ def extend_schema(
 
         - ``Serializer`` class
         - ``Serializer`` instance (e.g. ``Serializer(many=True)`` for listings)
-        - ``dict`` with status codes as keys and `Serializers` as values.
-        - ``dict`` with tuple (status_code, media_type) as keys and `Serializers` as values.
+        - basic types or instances of ``OpenApiTypes``
+        - :class:`.OpenApiResponse` for bundling any of the other choices together with
+          either a dedicated response description and/or examples.
         - :class:`.PolymorphicProxySerializer` for signaling that
           the operation may yield data from different serializers depending
           on the circumstances.
-    :param request: replaces the discovered ``Serializer``.
-    :param auth:
+        - ``dict`` with status codes as keys and one of the above as values.
+          Additionally in this case, it is also possible to provide a raw schema dict
+          as value.
+        - ``dict`` with tuples (status_code, media_type) as keys and one of the above
+          as values. Additionally in this case, it is also possible to provide a raw
+          schema dict as value.
+    :param request: replaces the discovered ``Serializer``. Takes a variety of inputs
+
+        - ``Serializer`` class/instance
+        - basic types or instances of ``OpenApiTypes``
+        - :class:`.PolymorphicProxySerializer` for signaling that the operation
+          accepts a set of different types of objects.
+        - ``dict`` with media_type as keys and one of the above as values. Additionally in
+          this case, it is also possible to provide a raw schema dict as value.
+    :param auth: replace discovered auth with explicit list of auth methods
     :param description: replaces discovered doc strings
     :param summary: an optional short summary of the description
     :param deprecated: mark operation as deprecated
     :param tags: override default list of tags
+    :param filters: ignore list detection and forcefully enable/disable filter discovery
     :param exclude: set True to exclude operation from schema
     :param operation: manually override what auto-discovery would generate. you must
         provide a OpenAPI3-compliant dictionary that gets directly translated to YAML.
     :param methods: scope extend_schema to specific methods. matches all by default.
     :param versions: scope extend_schema to specific API version. matches all by default.
     :param examples: attach request/response examples to the operation
+    :param extensions: specification extensions, e.g. ``x-badges``, ``x-code-samples``, etc.
     :return:
     """
+    if methods is not None:
+        methods = [method.upper() for method in methods]
+
     def decorator(f):
         BaseSchema = (
             # explicit manually set schema or previous view annotation
@@ -201,14 +288,14 @@ def extend_schema(
             return method_scope and version_scope
 
         class ExtendedSchema(BaseSchema):
-            def get_operation(self, path, path_regex, method, registry):
-                self.method = method
+            def get_operation(self, path, path_regex, path_prefix, method, registry):
+                self.method = method.upper()
 
                 if exclude and is_in_scope(self):
                     return None
                 if operation is not None and is_in_scope(self):
                     return operation
-                return super().get_operation(path, path_regex, method, registry)
+                return super().get_operation(path, path_regex, path_prefix, method, registry)
 
             def get_operation_id(self):
                 if operation_id and is_in_scope(self):
@@ -221,13 +308,13 @@ def extend_schema(
                 return super().get_override_parameters()
 
             def get_auth(self):
-                if auth and is_in_scope(self):
+                if auth is not None and is_in_scope(self):
                     return auth
                 return super().get_auth()
 
             def get_examples(self):
                 if examples and is_in_scope(self):
-                    return examples
+                    return super().get_examples() + examples
                 return super().get_examples()
 
             def get_request_serializer(self):
@@ -260,6 +347,16 @@ def extend_schema(
                     return tags
                 return super().get_tags()
 
+            def get_extensions(self):
+                if extensions and is_in_scope(self):
+                    return extensions
+                return super().get_extensions()
+
+            def get_filter_backends(self):
+                if filters is not None and is_in_scope(self):
+                    return getattr(self.view, 'filter_backends', []) if filters else []
+                return super().get_filter_backends()
+
         if inspect.isclass(f):
             # either direct decoration of views, or unpacked @api_view from OpenApiViewExtension
             if operation_id is not None or operation is not None:
@@ -269,11 +366,13 @@ def extend_schema(
                 )
             # reorder schema class MRO so that view method annotation takes precedence
             # over view class annotation. only relevant if there is a method annotation
-            for view_method in get_view_methods(view=f, schema=BaseSchema):
-                if 'schema' in getattr(view_method, 'kwargs', {}):
-                    view_method.kwargs['schema'] = type(
-                        'ExtendedMetaSchema', (view_method.kwargs['schema'], ExtendedSchema), {}
-                    )
+            for view_method_name in get_view_method_names(view=f, schema=BaseSchema):
+                if 'schema' not in getattr(getattr(f, view_method_name), 'kwargs', {}):
+                    continue
+                view_method = isolate_view_method(f, view_method_name)
+                view_method.kwargs['schema'] = type(
+                    'ExtendedMetaSchema', (view_method.kwargs['schema'], ExtendedSchema), {}
+                )
             # persist schema on class to provide annotation to derived view methods.
             # the second purpose is to serve as base for view multi-annotation
             f.schema = ExtendedSchema()
@@ -302,7 +401,10 @@ def extend_schema(
     return decorator
 
 
-def extend_schema_field(field, component_name=None):
+def extend_schema_field(
+        field: Union[_SerializerType, _FieldType, OpenApiTypes, Dict],
+        component_name: Optional[str] = None
+) -> Callable[[F], F]:
     """
     Decorator for the "field" kind. Can be used with ``SerializerMethodField`` (annotate the actual
     method) or with custom ``serializers.Field`` implementations.
@@ -325,77 +427,87 @@ def extend_schema_field(field, component_name=None):
 
 
 def extend_schema_serializer(
-        many=None,
-        exclude_fields=None,
+        many: Optional[bool] = None,
+        exclude_fields: Optional[List[str]] = None,
+        deprecate_fields: Optional[List[str]] = None,
         examples: Optional[List[OpenApiExample]] = None,
-):
+        extensions: Optional[Dict[str, Any]] = None,
+        component_name: Optional[str] = None,
+) -> Callable[[F], F]:
     """
     Decorator for the "serializer" kind. Intended for overriding default serializer behaviour that
-    cannot be influenced through `.extend_schema`.
+    cannot be influenced through :func:`@extend_schema <.extend_schema>`.
 
     :param many: override how serializer is initialized. Mainly used to coerce the list view detection
         heuristic to acknowledge a non-list serializer.
     :param exclude_fields: fields to ignore while processing the serializer. only affects the
         schema. fields will still be exposed through the API.
+    :param deprecate_fields: fields to mark as deprecated while processing the serializer.
     :param examples: define example data to serializer.
+    :param extensions: specification extensions, e.g. ``x-is-dynamic``, etc.
+    :param component_name: override default class name extraction.
     """
     def decorator(klass):
         if many is not None:
             set_override(klass, 'many', many)
         if exclude_fields:
             set_override(klass, 'exclude_fields', exclude_fields)
+        if deprecate_fields:
+            set_override(klass, 'deprecate_fields', deprecate_fields)
         if examples:
             set_override(klass, 'examples', examples)
+        if extensions:
+            set_override(klass, 'extensions', extensions)
+        if component_name:
+            set_override(klass, 'component_name', component_name)
         return klass
 
     return decorator
 
 
-def extend_schema_view(**kwargs):
+def extend_schema_view(**kwargs) -> Callable[[F], F]:
     """
     Convenience decorator for the "view" kind. Intended for annotating derived view methods that
     are are not directly present in the view (usually methods like ``list`` or ``retrieve``).
     Spares you from overriding methods like ``list``, only to perform a super call in the body
-    so that you have have something to attach ``@extend_schema`` to.
+    so that you have have something to attach :func:`@extend_schema <.extend_schema>` to.
 
-    :param kwargs: method names as argument names and ``extend_schema()`` calls as values
+    This decorator also takes care of safely attaching annotations to derived view methods,
+    preventing leakage into unrelated views.
+
+    :param kwargs: method names as argument names and :func:`@extend_schema <.extend_schema>`
+      calls as values
     """
-    def wrapping_decorator(method_decorator, method):
-        @method_decorator
-        @functools.wraps(method)
-        def wrapped_method(self, request, *args, **kwargs):
-            return method(self, request, *args, **kwargs)
-
-        return wrapped_method
-
     def decorator(view):
-        view_methods = {m.__name__: m for m in get_view_methods(view)}
+        # special case for @api_view. redirect decoration to enclosed WrappedAPIView
+        if callable(view) and hasattr(view, 'cls'):
+            extend_schema_view(**kwargs)(view.cls)
+            return view
+
+        available_view_methods = get_view_method_names(view)
 
         for method_name, method_decorator in kwargs.items():
-            if method_name not in view_methods:
+            if method_name not in available_view_methods:
                 warn(
                     f'@extend_schema_view argument "{method_name}" was not found on view '
                     f'{view.__name__}. method override for "{method_name}" will be ignored.'
                 )
                 continue
 
-            method = view_methods[method_name]
-            # the context of derived methods must not be altered, as it belongs to the other
-            # class. create a new context via the wrapping_decorator so the schema can be safely
-            # stored in the wrapped_method. methods belonging to the view can be safely altered.
-            if method_name in view.__dict__:
-                method_decorator(method)
-            else:
-                setattr(view, method_name, wrapping_decorator(method_decorator, method))
+            # the context of derived methods must not be altered, as it belongs to the
+            # other view. create a new context so the schema can be safely stored in the
+            # wrapped_method. view methods that are not derived can be safely altered.
+            method_decorator(isolate_view_method(view, method_name))
         return view
 
     return decorator
 
 
-def inline_serializer(name: str, fields: Dict[str, object], **kwargs) -> SerializerType:
+def inline_serializer(name: str, fields: Dict[str, Field], **kwargs) -> Serializer:
     """
-    A helper function to create an inline serializer. Primary use is with `@extend_schema`,
-    where one needs an implicit one-off serializer that is not reflected in an actual class.
+    A helper function to create an inline serializer. Primary use is with
+    :func:`@extend_schema <.extend_schema>`, where one needs an implicit one-off
+    serializer that is not reflected in an actual class.
 
     :param name: name of the
     :param fields: dict with field names as keys and serializer fields as values
