@@ -18,12 +18,12 @@ def pytest_configure(config):
         'dj_rest_auth.registration',
         'allauth',
         'allauth.account',
-        'rest_framework_jwt',
         'oauth2_provider',
         'django_filters',
         # this is not strictly required and when added django-polymorphic
         # currently breaks the whole Django/DRF upstream testing.
         # 'polymorphic',
+        # 'rest_framework_jwt',
     ]
 
     base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
@@ -89,7 +89,12 @@ def pytest_configure(config):
         REST_FRAMEWORK={
             'DEFAULT_SCHEMA_CLASS': 'drf_spectacular.openapi.AutoSchema',
             'PAGE_SIZE': 10,
-        }
+        },
+        DEFAULT_AUTO_FIELD='django.db.models.AutoField',
+        SILENCED_SYSTEM_CHECKS=[
+            'rest_framework.W001', 'fields.E210', 'security.W001', 'security.W002',
+            'security.W003', 'security.W009', 'security.W012'
+        ],
     )
 
     django.setup()
@@ -120,7 +125,7 @@ def pytest_collection_modifyitems(config, items):
     for item in items:
         for marker in item.own_markers:
             if marker.name == 'contrib' and config.getoption("--skip-missing-contrib"):
-                if not all([module_available(module_str) for module_str in marker.args]):
+                if not all([module_available(module_str) for module_str in marker.args]):  # pragma: no cover
                     item.add_marker(skip_missing_contrib)
             if marker.name == 'contrib' and config.getoption("--allow-contrib-fail"):
                 item.add_marker(allow_contrib_fail)
@@ -143,6 +148,16 @@ def warnings(capsys):
     assert captured.err
 
 
+@pytest.fixture()
+def clear_caches():
+    from drf_spectacular.plumbing import get_openapi_type_mapping, load_enum_name_overrides
+    load_enum_name_overrides.cache_clear()
+    get_openapi_type_mapping.cache_clear()
+    yield
+    load_enum_name_overrides.cache_clear()
+    get_openapi_type_mapping.cache_clear()
+
+
 def module_available(module_str):
     try:
         import_module(module_str)
