@@ -7,7 +7,9 @@ from rest_framework.response import Response
 
 from drf_spectacular.openapi import AutoSchema
 from drf_spectacular.types import OpenApiTypes
-from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_field
+from drf_spectacular.utils import (
+    OpenApiParameter, extend_schema, extend_schema_field, extend_schema_serializer,
+)
 from tests import assert_schema, generate_schema, get_response_schema
 
 
@@ -31,6 +33,7 @@ class CustomField(serializers.Field):
         return urlsafe_base64_encode(b'\xf0\xf1\xf2')  # pragma: no cover
 
 
+@extend_schema_serializer(component_name='GammaEpsilon')
 class GammaSerializer(serializers.Serializer):
     encoding = serializers.CharField()
     image_data = CustomField()
@@ -50,6 +53,7 @@ class QuerySerializer(serializers.Serializer):
         choices=['a', 'b', 'c'],
         default=['a'],
     )
+    tag = serializers.CharField(required=False)
 
 
 class ErrorDetailSerializer(serializers.Serializer):
@@ -57,6 +61,7 @@ class ErrorDetailSerializer(serializers.Serializer):
     field_j = serializers.SerializerMethodField()
     field_k = serializers.SerializerMethodField()
     field_l = serializers.SerializerMethodField()
+    field_m = serializers.SerializerMethodField()
 
     @extend_schema_field(OpenApiTypes.DATETIME)
     def get_field_i(self, object):
@@ -74,6 +79,10 @@ class ErrorDetailSerializer(serializers.Serializer):
     def get_field_l(self, object):
         return object.some_choice  # pragma: no cover
 
+    @extend_schema_field({'type': 'array', 'items': {'type': 'integer'}})
+    def get_field_m(self, object):
+        return [1, 2, 3]  # pragma: no cover
+
 
 with mock.patch('rest_framework.settings.api_settings.DEFAULT_SCHEMA_CLASS', AutoSchema):
     class DoesItAllViewset(viewsets.GenericViewSet):
@@ -92,9 +101,26 @@ with mock.patch('rest_framework.settings.api_settings.DEFAULT_SCHEMA_CLASS', Aut
                     'expiration_date', OpenApiTypes.DATETIME, description='time the object will expire at'
                 ),
                 OpenApiParameter(
-                    'test_mode', bool, location=OpenApiParameter.HEADER, enum=[True, False],
+                    name='test_mode',
+                    type=bool,
+                    location=OpenApiParameter.HEADER,
+                    enum=[True, False],
+                    default=False,
                     description='creation will be in the sandbox',
-                )
+                ),
+                OpenApiParameter(
+                    name='X-Api-Version',
+                    type=str,
+                    location=OpenApiParameter.HEADER,
+                    response=True,
+                ),
+                OpenApiParameter(
+                    name='Location',
+                    type=OpenApiTypes.URI,
+                    location=OpenApiParameter.HEADER,
+                    description='URL of the created resource',
+                    response=[201],
+                ),
             ],
             description='this weird endpoint needs some explaining',
             summary='short summary',
@@ -172,6 +198,18 @@ with mock.patch('rest_framework.settings.api_settings.DEFAULT_SCHEMA_CLASS', Aut
         def non_required_body(self, request):
             return Response([])  # pragma: no cover
 
+        @extend_schema(
+            request={
+                'application/json': dict,
+                'application/pdf': bytes,
+                'text/html': OpenApiTypes.STR
+            },
+            responses=None
+        )
+        @action(detail=False, methods=['POST'])
+        def custom_request_override(self, request):
+            return Response([])  # pragma: no cover
+
         @extend_schema(responses={
             (200, 'application/pdf'): bytes
         })
@@ -322,4 +360,25 @@ def test_extend_schema_field_custom_schema_with_without_breakout(no_warnings):
     assert properties['field']['description'] == 'some explaining'
     assert properties['field']['readOnly']
     assert 'Breakout' in properties['field_breakout']['allOf'][0]['$ref']
-    properties['field_breakout']['readOnly']
+    assert properties['field_breakout']['readOnly']
+
+
+def test_extend_schema_field_with_field_class(no_warnings) -> None:
+    class XSerializer(serializers.Serializer):
+        field = serializers.SerializerMethodField()
+
+        @extend_schema_field(serializers.IntegerField())
+        def get_field(self, object):
+            return 1  # pragma: no cover
+
+    @extend_schema(responses=XSerializer)
+    @api_view(['GET'])
+    def view_func(request, format=None):
+        pass  # pragma: no cover
+
+    schema = generate_schema('x', view_function=view_func)
+    assert schema['components']['schemas']['X'] == {
+        'type': 'object',
+        'properties': {'field': {'type': 'integer', 'readOnly': True}},
+        'required': ['field']
+    }
