@@ -1,7 +1,11 @@
+import functools
+import json
+import sys
 import tempfile
 import uuid
-from datetime import date, datetime, timedelta
+from datetime import timedelta
 from decimal import Decimal
+from typing import Optional
 
 import pytest
 from django import __version__ as DJANGO_VERSION
@@ -15,7 +19,15 @@ from rest_framework.routers import SimpleRouter
 from rest_framework.test import APIClient
 
 from drf_spectacular.generators import SchemaGenerator
-from tests import assert_schema
+from tests import assert_equal, assert_schema, build_absolute_file_path
+
+if sys.version_info >= (3, 8):
+    functools_cached_property = functools.cached_property
+else:
+    # We re-use Django's cached_property when it's not avaiable to
+    # keep tests unified across Python versions.
+    functools_cached_property = cached_property
+
 
 fs = FileSystemStorage(location=tempfile.gettempdir())
 
@@ -32,22 +44,39 @@ class AuxSerializer(serializers.ModelSerializer):
         model = Aux
 
 
+class SubObject:
+    def __init__(self, instance):
+        self._instance = instance
+
+    @property
+    def calculated(self) -> int:
+        return self._instance.field_int
+
+    @property
+    def nested(self) -> 'SubObject':
+        return self
+
+    @property
+    def model_instance(self) -> 'AllFields':
+        return self._instance
+
+    @property
+    def optional_int(self) -> Optional[int]:
+        return 1
+
+
 class AllFields(models.Model):
     # basics
     field_int = models.IntegerField()
     field_float = models.FloatField()
     field_bool = models.BooleanField()
     field_char = models.CharField(max_length=100)
-    field_text = models.TextField()
+    field_text = models.TextField(verbose_name='a text field')
     # special
     field_slug = models.SlugField()
     field_email = models.EmailField()
     field_uuid = models.UUIDField()
     field_url = models.URLField()
-    if models.IPAddressField in serializers.ModelSerializer.serializer_field_mapping:
-        field_ip = models.IPAddressField()
-    else:
-        field_ip = models.GenericIPAddressField(protocol='ipv6')
     field_ip_generic = models.GenericIPAddressField(protocol='ipv6')
     field_decimal = models.DecimalField(max_digits=6, decimal_places=3)
     field_file = models.FileField(storage=fs)
@@ -58,14 +87,24 @@ class AllFields(models.Model):
     field_smallint = models.SmallIntegerField()
     field_posint = models.PositiveIntegerField()
     field_possmallint = models.PositiveSmallIntegerField()
-    field_nullbool = models.NullBooleanField()
+    if DJANGO_VERSION > '3.1':
+        field_nullbool = models.BooleanField(null=True)
+    else:
+        field_nullbool = models.NullBooleanField()  # type: ignore
     field_time = models.TimeField()
     field_duration = models.DurationField()
+    field_binary = models.BinaryField()
 
     # relations
-    field_foreign = models.ForeignKey(Aux, on_delete=models.CASCADE, help_text='main aux object')
-    field_m2m = models.ManyToManyField(Aux, help_text='set of related aux objects')
-    field_o2o = models.OneToOneField(Aux, on_delete=models.CASCADE, help_text='bound aux object')
+    field_foreign = models.ForeignKey(
+        Aux, on_delete=models.CASCADE, help_text='main aux object', related_name='ff'
+    )
+    field_m2m = models.ManyToManyField(
+        Aux, help_text='set of related aux objects', related_name='fm'
+    )
+    field_o2o = models.OneToOneField(
+        Aux, on_delete=models.CASCADE, help_text='bound aux object', related_name='fo'
+    )
     # overrides
     field_regex = models.CharField(max_length=50)
     field_bool_override = models.BooleanField()
@@ -85,6 +124,10 @@ class AllFields(models.Model):
     def field_model_cached_property_float(self) -> float:
         return 1.337
 
+    @functools_cached_property
+    def field_model_py_cached_property_float(self) -> float:
+        return 1.337
+
     @property
     def field_list(self):
         return [1.1, 2.2, 3.3]
@@ -99,6 +142,22 @@ class AllFields(models.Model):
     def model_function_model(self) -> Aux:
         return self.field_foreign
 
+    @property
+    def sub_object(self) -> SubObject:
+        return SubObject(self)
+
+    @cached_property
+    def sub_object_cached(self) -> SubObject:
+        return SubObject(self)
+
+    @functools_cached_property
+    def sub_object_py_cached(self) -> SubObject:
+        return SubObject(self)
+
+    @property
+    def optional_sub_object(self) -> Optional[SubObject]:
+        return SubObject(self)
+
 
 class AllFieldsSerializer(serializers.ModelSerializer):
     field_decimal_uncoerced = serializers.DecimalField(
@@ -117,7 +176,9 @@ class AllFieldsSerializer(serializers.ModelSerializer):
     def get_field_method_object(self, obj) -> dict:
         return {'key': 'value'}
 
-    field_regex = serializers.RegexField(r'^[a-zA-z0-9]{10}\-[a-z]')
+    field_regex = serializers.RegexField(r'^[a-zA-z0-9]{10}\-[a-z]', label='A regex field')
+
+    field_hidden = serializers.HiddenField(default='')
 
     # composite fields
     field_list = serializers.ListField(
@@ -131,13 +192,13 @@ class AllFieldsSerializer(serializers.ModelSerializer):
     # extra related fields
     field_related_slug = serializers.SlugRelatedField(
         read_only=True, source='field_foreign', slug_field='id'
-    )
+    )  # type: ignore
     field_related_string = serializers.StringRelatedField(
         source='field_foreign'
-    )
+    )  # type: ignore
     field_related_hyperlink = serializers.HyperlinkedRelatedField(
         read_only=True, source='field_foreign', view_name='aux-detail'
-    )
+    )  # type: ignore
     field_identity_hyperlink = serializers.HyperlinkedIdentityField(
         read_only=True, view_name='allfields-detail'
     )
@@ -158,6 +219,8 @@ class AllFieldsSerializer(serializers.ModelSerializer):
 
     field_model_cached_property_float = serializers.ReadOnlyField()
 
+    field_model_py_cached_property_float = serializers.ReadOnlyField()
+
     field_dict_int = serializers.DictField(
         child=serializers.IntegerField(),
         source='field_json',
@@ -167,6 +230,33 @@ class AllFieldsSerializer(serializers.ModelSerializer):
     # need to set the field explicitly. defined here for both cases to have consistent ordering.
     field_json = serializers.JSONField()
 
+    # traversal of non-model types of complex object
+    field_sub_object_calculated = serializers.ReadOnlyField(source='sub_object.calculated')
+    field_sub_object_nested_calculated = serializers.ReadOnlyField(source='sub_object.nested.calculated')
+    field_sub_object_model_int = serializers.ReadOnlyField(source='sub_object.model_instance.field_int')
+
+    field_sub_object_cached_calculated = serializers.ReadOnlyField(source='sub_object_cached.calculated')
+    field_sub_object_cached_nested_calculated = serializers.ReadOnlyField(source='sub_object_cached.nested.calculated')
+    field_sub_object_cached_model_int = serializers.ReadOnlyField(source='sub_object_cached.model_instance.field_int')
+
+    field_sub_object_py_cached_calculated = serializers.ReadOnlyField(source='sub_object_py_cached.calculated')
+    field_sub_object_py_cached_nested_calculated = serializers.ReadOnlyField(
+        source='sub_object_py_cached.nested.calculated',
+    )
+    field_sub_object_py_cached_model_int = serializers.ReadOnlyField(
+        source='sub_object_py_cached.model_instance.field_int',
+    )
+
+    # typing.Optional
+    field_optional_sub_object_calculated = serializers.ReadOnlyField(
+        source='optional_sub_object.calculated',
+        allow_null=True,
+    )
+    field_sub_object_optional_int = serializers.ReadOnlyField(
+        source='sub_object.optional_int',
+        allow_null=True,
+    )
+
     class Meta:
         fields = '__all__'
         model = AllFields
@@ -202,7 +292,7 @@ def test_fields(no_warnings):
 @pytest.mark.urls(__name__)
 @pytest.mark.django_db
 def test_model_setup_is_valid():
-    aux = Aux()
+    aux = Aux(id='0ac6930d-87f4-40e8-8242-10a3ed31a335')
     aux.save()
 
     m = AllFields(
@@ -217,13 +307,12 @@ def test_model_setup_is_valid():
         field_email='test@example.com',
         field_uuid='00000000-00000000-00000000-00000000',
         field_url='https://github.com/tfranzel/drf-spectacular',
-        field_ip='127.0.0.1',
         field_ip_generic='2001:db8::8a2e:370:7334',
         field_decimal=Decimal('666.333'),
         field_file=None,
         field_img=None,  # TODO fill with data below
-        field_date=date.today(),
-        field_datetime=datetime.now(),
+        field_date='2021-09-09',
+        field_datetime='2021-09-09T10:15:26.049862',
         field_bigint=11111111111111,
         field_smallint=111111,
         field_posint=123,
@@ -231,6 +320,7 @@ def test_model_setup_is_valid():
         field_nullbool=None,
         field_time='00:05:23.283',
         field_duration=timedelta(seconds=10),
+        field_binary=b'\xce\x9c\xce\xb9\xce\xb1',
         # relations
         field_foreign=aux,
         field_o2o=aux,
@@ -246,3 +336,13 @@ def test_model_setup_is_valid():
 
     response = APIClient().get(reverse('allfields-detail', args=(m.pk,)))
     assert response.status_code == 200
+
+    with open(build_absolute_file_path('tests/test_fields_response.json')) as fh:
+        expected = json.load(fh)
+
+    if DJANGO_VERSION < '3':
+        expected['field_file'] = f'http://testserver/allfields/1/{m.field_file.name}'
+    else:
+        expected['field_file'] = f'http://testserver/{m.field_file.name}'
+
+    assert_equal(json.loads(response.content), expected)
