@@ -1,3 +1,4 @@
+from typing import Union
 from unittest import mock
 
 import pytest
@@ -9,9 +10,10 @@ from rest_framework.decorators import action, api_view
 from rest_framework.schemas import AutoSchema as DRFAutoSchema
 from rest_framework.views import APIView
 
-from drf_spectacular.generators import SchemaGenerator
 from drf_spectacular.types import OpenApiTypes
-from drf_spectacular.utils import PolymorphicProxySerializer, extend_schema, extend_schema_view
+from drf_spectacular.utils import (
+    OpenApiParameter, PolymorphicProxySerializer, extend_schema, extend_schema_view,
+)
 from tests import generate_schema
 from tests.models import SimpleModel, SimpleSerializer
 
@@ -19,7 +21,6 @@ from tests.models import SimpleModel, SimpleSerializer
 def test_serializer_name_reuse(capsys):
     from rest_framework import routers
 
-    from drf_spectacular.generators import SchemaGenerator
     router = routers.SimpleRouter()
 
     def x1():
@@ -44,8 +45,7 @@ def test_serializer_name_reuse(capsys):
 
     router.register('x2', X2Viewset, basename='x2')
 
-    generator = SchemaGenerator(patterns=router.urls)
-    generator.get_schema(request=None, public=True)
+    generate_schema(None, patterns=router.urls)
 
     stderr = capsys.readouterr().err
     assert 'Encountered 2 components with identical names "X" and different classes' in stderr
@@ -78,8 +78,10 @@ def test_no_queryset_warn(capsys):
 
     generate_schema('x1', X1Viewset)
     stderr = capsys.readouterr().err
-    assert 'obtaining queryset from' in stderr  # warning 1
-    assert 'failed to obtain model through view\'s queryset' in stderr  # warning 2
+    assert (
+        'could not derive type of path parameter "id" because it '
+        'is untyped and obtaining queryset from the viewset failed.'
+    ) in stderr
 
 
 def test_path_param_not_in_model(capsys):
@@ -92,7 +94,9 @@ def test_path_param_not_in_model(capsys):
             pass  # pragma: no cover
 
     generate_schema('x1', XViewset)
-    assert 'no such field' in capsys.readouterr().err
+    stderr = capsys.readouterr().err
+    assert 'no such field' in stderr
+    assert 'XViewset' in stderr
 
 
 def test_no_authentication_scheme_registered(capsys):
@@ -107,7 +111,9 @@ def test_no_authentication_scheme_registered(capsys):
         authentication_classes = [XAuth]
 
     generate_schema('x', XViewset)
-    assert 'no OpenApiAuthenticationExtension registered' in capsys.readouterr().err
+    stderr = capsys.readouterr().err
+    assert 'no OpenApiAuthenticationExtension registered' in stderr
+    assert 'XViewset' in stderr
 
 
 def test_serializer_not_found(capsys):
@@ -115,7 +121,7 @@ def test_serializer_not_found(capsys):
         pass  # pragma: no cover
 
     generate_schema('x', XViewset)
-    assert 'Exception raised while getting serializer' in capsys.readouterr().err
+    assert 'XViewset: exception raised while getting serializer.' in capsys.readouterr().err
 
 
 def test_extend_schema_unknown_class(capsys):
@@ -150,7 +156,7 @@ def test_no_serializer_class_on_apiview(capsys):
             pass  # pragma: no cover
 
     generate_schema('x', view=XView)
-    assert 'Unable to guess serializer for' in capsys.readouterr().err
+    assert 'XView: unable to guess serializer.' in capsys.readouterr().err
 
 
 def test_unable_to_follow_field_source_through_intermediate_property_warning(capsys):
@@ -173,7 +179,7 @@ def test_unable_to_follow_field_source_through_intermediate_property_warning(cap
 
     generate_schema('x', view=XAPIView)
     assert (
-        'could not follow field source through intermediate property'
+        'XAPIView: XSerializer: could not follow field source through intermediate property'
     ) in capsys.readouterr().err
 
 
@@ -201,8 +207,39 @@ def test_unable_to_derive_function_type_warning(capsys):
 
     generate_schema('x', view=XAPIView)
     stderr = capsys.readouterr().err
-    assert 'unable to resolve type hint for function "x"' in stderr
-    assert 'unable to resolve type hint for function "get_y"' in stderr
+    assert 'XAPIView: XSerializer: unable to resolve type hint for function "x"' in stderr
+    assert 'XAPIView: XSerializer: unable to resolve type hint for function "get_y"' in stderr
+
+
+def test_unable_to_traverse_union_type_hint(capsys):
+    class Foo:
+        foo_value: int = 1
+
+    class Bar:
+        pass
+
+    class FailingFieldSourceTraversalModel3(models.Model):
+        @property
+        def foo_or_bar(self) -> Union[Foo, Bar]:
+            pass  # pragma: no cover
+
+    class XSerializer(serializers.ModelSerializer):
+        foo_value = serializers.ReadOnlyField(source='foo_or_bar.foo_value')
+
+        class Meta:
+            model = FailingFieldSourceTraversalModel3
+            fields = '__all__'
+
+    class XAPIView(APIView):
+        @extend_schema(responses=XSerializer)
+        def get(self, request):
+            pass  # pragma: no cover
+
+    generate_schema('foo_value', view=XAPIView)
+    stderr = capsys.readouterr().err
+    assert 'could not traverse Union type' in stderr
+    assert 'Foo' in stderr
+    assert 'Bar' in stderr
 
 
 def test_operation_id_collision_resolution(capsys):
@@ -215,8 +252,7 @@ def test_operation_id_collision_resolution(capsys):
         path('pi/<int:foo>', view_func),
         path('pi/', view_func),
     ]
-    generator = SchemaGenerator(patterns=urlpatterns)
-    schema = generator.get_schema(request=None, public=True)
+    schema = generate_schema(None, patterns=urlpatterns)
 
     assert schema['paths']['/pi/']['get']['operationId'] == 'pi_retrieve'
     assert schema['paths']['/pi/{foo}']['get']['operationId'] == 'pi_retrieve_2'
@@ -291,3 +327,109 @@ def test_warning_request_body_not_resolvable(capsys):
     generate_schema('x', view_function=view_func)
     stderr = capsys.readouterr().err
     assert 'could not resolve request body for POST /x' in stderr
+
+
+def test_response_header_warnings(capsys):
+    @extend_schema(
+        request=OpenApiTypes.ANY,
+        responses=OpenApiTypes.ANY,
+        parameters=[OpenApiParameter(name='test', response=True)]
+    )
+    @api_view(['POST'])
+    def view_func(request, format=None):
+        pass  # pragma: no cover
+
+    generate_schema('x', view_function=view_func)
+    stderr = capsys.readouterr().err
+    assert 'incompatible location type ignored' in stderr
+
+
+def test_unknown_base_field_warning(capsys):
+    class UnknownField(serializers.Field):
+        pass
+
+    class XSerializer(serializers.Serializer):
+        field = UnknownField()
+
+    @extend_schema(request=XSerializer, responses=XSerializer)
+    @api_view(['POST'])
+    def view_func(request, format=None):
+        pass  # pragma: no cover
+
+    generate_schema('x', view_function=view_func)
+    stderr = capsys.readouterr().err
+    assert 'could not resolve serializer field' in stderr
+
+
+def test_warning_read_only_field_on_non_model_serializer(capsys):
+    class XSerializer(serializers.Serializer):
+        field = serializers.ReadOnlyField()
+
+    class XViewSet(viewsets.ModelViewSet):
+        serializer_class = XSerializer
+        queryset = SimpleModel.objects.all()
+
+    # test validity of serializer construction
+    serializer = XSerializer(instance={'field': 1})
+    serializer.data
+
+    generate_schema('x', XViewSet)
+    stderr = capsys.readouterr().err
+    assert 'Could not derive type for ReadOnlyField "field"' in stderr
+
+
+def test_warning_missing_lookup_field_on_model_serializer(capsys):
+    class XViewSet(viewsets.ModelViewSet):
+        serializer_class = SimpleSerializer
+        queryset = SimpleModel.objects.all()
+        lookup_field = 'non_existent_field'
+
+    generate_schema('x', XViewSet)
+    stderr = capsys.readouterr().err
+    assert (
+        'could not derive type of path parameter "non_existent_field" because model '
+        '"tests.models.SimpleModel" contained no such field.'
+    ) in stderr
+
+
+@mock.patch(
+    'drf_spectacular.settings.spectacular_settings.PATH_CONVERTER_OVERRIDES', {'int': object}
+)
+def test_invalid_path_converter_override(capsys):
+    @extend_schema(responses=OpenApiTypes.FLOAT)
+    @api_view(['GET'])
+    def pi(request, foo):
+        pass  # pragma: no cover
+
+    urlpatterns = [path('/a/<int:var>/', pi)]
+    generate_schema(None, patterns=urlpatterns)
+    stderr = capsys.readouterr().err
+    assert 'Unable to use path converter override for "int".' in stderr
+
+
+def test_malformed_vendor_extensions(capsys):
+    @extend_schema(
+        responses=None,
+        extensions={'foo': 'not-prefixed'}
+    )
+    @api_view(['GET'])
+    def view_func(request, format=None):
+        pass  # pragma: no cover
+
+    generate_schema('x', view_function=view_func)
+    stderr = capsys.readouterr().err
+    assert 'invalid extension \'foo\'. vendor extensions must start with' in stderr
+
+
+def test_serializer_method_missing(capsys):
+    class XSerializer(serializers.Serializer):
+        some_field = serializers.SerializerMethodField()
+
+    @extend_schema(responses=XSerializer)
+    @api_view(['GET'])
+    def view_func(request):
+        pass  # pragma: no cover
+
+    generate_schema('x', view_function=view_func)
+    stderr = capsys.readouterr().err
+    assert 'SerializerMethodField "some_field" is missing required method "get_some_field"' in stderr
